#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon May 18 14:38:41 2020

@author: xiao
"""

import os, glob, gc, time, shutil, numpy as np

from pathlib import Path
import h5py, tifffile, json
from scipy.ndimage import zoom, gaussian_filter as gf, median_filter as median
import skimage.restoration as skr
from skimage.transform import rescale
import tomopy
from .io import (data_reader, tomo_h5_reader,
                 data_info, tomo_h5_info)

IF_LOG = True

TOMO_RECON_PARAM_DICT = {"file_params": {"raw_data_top_dir": None,
                                         "data_center_dir": None,
                                         "recon_top_dir": None,
                                         "debug_top_dir": None,
                                         "cen_list_file": None,
                                         "alt_flat_file": None,
                                         "alt_dark_file": None,
                                         "wedge_ang_auto_det_ref_fn": None},
                         "recon_config": {"recon_type": 'Trial Cent',
                                          "use_debug": False,
                                          "use_alt_flat": False,
                                          "use_alt_dark": False,
                                          "use_fake_flat": False,
                                          "use_fake_dark": False,
                                          "use_rm_zinger": False,
                                          "use_mask": True,
                                          "use_wedge_ang_auto_det": False,
                                          "is_wedge": False}, "flt_params": {"filters": {}},
                         "data_params": {"scan_id": 0,
                                         "downsample": 1,
                                         "rot_cen": 1280,
                                         "cen_win_s": 1240,
                                         "cen_win_w": 8,
                                         "fake_flat_val": 1e4,
                                         "fake_dark_val": 1e2,
                                         "sli_s": 1280,
                                         "sli_e": 1300,
                                         "chunk_sz": 200,
                                         "margin": 15,
                                         "zinger_val": 500,
                                         "mask_ratio": 1,
                                         # "wedge_blankat":90,
                                         "wedge_missing_s": 500,
                                         "wedge_missing_e": 600,
                                         "wedge_ang_auto_det_thres": 0.1}, "alg_params": {}}

FILTERLIST = ["phase retrieval",
              "flatting bkg",
              "remove cupping",
              "stripe_removal: vo",
              "stripe_removal: ti",
              "stripe_removal: sf",
              "stripe_removal: fw",
              "denoise: median",
              "denoise: wiener",
              "denoise: unsupervised_wiener",
              "denoise: denoise_nl_means",
              "denoise: denoise_tv_bregman",
              "denoise: denoise_tv_chambolle",
              "denoise: denoise_bilateral",
              "denoise: denoise_wavelet"]


def align_proj(data, data_ref=None, **kwargs):
    pass


def data_down_sampling(data, levels):
    if np.any((np.array(levels) - 1) > 0):
        return rescale(data, 1 / np.array(levels), mode='edge', anti_aliasing=True)
    else:
        return rescale(data, 1 / np.array(levels), mode='edge', anti_aliasing=False)


def get_dim_ang_range_range_sli(dim_info, reader, fn, cfg,
                                sli_start=0, sli_end=None,
                                col_start=None, col_end=None,
                                ds_use=False, ds_level=1.0, thres=0.1):
    """
    read_data(reader, fn, cfg, sli_start=0, sli_end=20,
              col_start=None, col_end=None,
              flat_name=None, dark_name=None,
              use_fake_flat=False, use_fake_dark=False,
              fake_flat_val=1e4, fake_dark_val=100,
              ds_use=False, ds_level=1.0, mean_axis=None)

    inputs:
        fn: str,
            full file path to the h5 raw data file
        sli_start: int
            starting slice index
        chunk_sz: int, optional
            the slice range that has missing angle range same as that at a representative slice;
            if chunk_sz=None, the same missing angle range at sli_start will be applied to the
            entire slice rangle; otherwise, the missing angle range will be computed chunk by chunk
        sli_end: int, optional
            if sli_end=None, the sli_end=slice.max;
        dataset_path: str, optional
            the path to the tomo dataset in h5 file structure; default is '/img_tomo'
        thres: float, optional
            count level below which is regarded as bad; default is 500
        blcok_view_at: int
            it indicates the missing angle range; it can only take two values: 0, or 90.
            if it is 0, the missing angles are in the beginning and end of 180 scans;
            if it is 90, the missing angles are in the middle of 180 scans.

    return:
        dictionary,
        the bad angle list for each slice in the slice range [sli_start, sli_end]
    """
    sli_start = int(sli_start)

    img_shape = dim_info(fn, dtype='data', cfg=cfg)
    if sli_end is None:
        sli_end = img_shape[1]
    elif sli_end > img_shape[1]:
        print('sli_end exceeds the maximum allowed range...')
        return None
    else:
        sli_end = int(sli_end)

    if col_start is None:
        col_start = 0
    else:
        col_start = int(col_start)

    if col_end is None:
        col_end = img_shape[2]
    else:
        col_end = int(col_end)

    tem, _, _, _ = read_data(reader, fn, cfg,
                             sli_start=sli_start, sli_end=sli_end,
                             col_start=col_start, col_end=col_end,
                             flat_name=None, dark_name=None,
                             use_fake_flat=False, use_fake_dark=False,
                             fake_flat_val=1e4, fake_dark_val=100,
                             ds_use=ds_use, ds_level=ds_level, mean_axis=2)
    bad_angs = tem < thres
    return bad_angs


def get_file(raw_data_top_dir, scan_id, cfg, recon_top_dir=None):
    print(cfg['tomo_raw_fn_template'].format(scan_id))
    data_file = glob.glob(os.path.join(raw_data_top_dir, cfg['tomo_raw_fn_template'].format(scan_id)))

    if data_file is []:
        return None

    if recon_top_dir is None:
        output_file = os.path.join(raw_data_top_dir,
                                   'recon_' + os.path.basename(data_file[0]).split(".")[-2],
                                   'recon_' + os.path.basename(data_file[0]).split(".")[-2] + "_{0}.tiff")
    else:
        output_file = os.path.join(recon_top_dir,
                                   'recon_' + os.path.basename(data_file[0]).split(".")[-2],
                                   'recon_' + os.path.basename(data_file[0]).split(".")[-2] + "_{0}.tiff")
    if (not os.path.exists(os.path.dirname(output_file))) and (recon_top_dir is not None):
        os.makedirs(os.path.dirname(output_file), mode=0o777)
    return data_file[0], output_file


def if_log(flt_dict):
    if sorted(flt_dict.keys())[0]:
        for key in sorted(flt_dict.keys()):
            if 'phase retrieval' == flt_dict[key]['filter_name']:
                if 'bronnikov' == flt_dict[key]['params']['filter']:
                    return False
        return True
    else:
        return True


def normalize(arr, flat, dark, fake_flat_roi=None, cutoff=None, ncore=None, out=None):
    """
    Normalize raw projection data using the flat and dark field projections.

    Parameters
    ----------
    arr : ndarray
        3D stack of projections.
    flat : ndarray
        3D flat field data.
    dark : ndarray
        3D dark field data.
    cutoff : float, optional
        Permitted maximum vaue for the normalized data.
    ncore : int, optional
        Number of cores that will be assigned to jobs.
    out : ndarray, optional
        Output array for result. If same as arr,
        process will be done in-place.

    Returns
    -------
    ndarray
        Normalized 3D tomographic data.
    """
    if fake_flat_roi is None:
        flat = np.mean(flat, axis=0, dtype=np.float32)
    dark = np.mean(dark, axis=0, dtype=np.float32)
    with tomopy.util.mproc.set_numexpr_threads(ncore):
        denom = (flat - dark).astype(np.float32)
        out = (arr - dark).astype(np.float32)
        out[:] = (out / denom)[:]
        out[np.isnan(out)] = 1
        out[np.isinf(out)] = 1
        out[out <= 0] = 1
        if cutoff is not None:
            cutoff = np.float32(cutoff)
            out[:] = np.where(out > cutoff, cutoff, out)[:]
    return out


def read_data(reader, fn, cfg, sli_start=0, sli_end=20,
              col_start=None, col_end=None,
              flat_name=None, dark_name=None,
              use_fake_flat=False, use_fake_dark=False,
              fake_flat_val=1e4, fake_dark_val=100,
              fake_flat_roi=None,
              ds_use=False, ds_level=1.0, mean_axis=None):
    """
    input:
        fn: str
            full file path to the h5 raw data file
        sli_start: int, optional
            starting slice index
        sli_end: int, optional
            if sli_end=None, the sli_end=slice.max;
        flat_name: str, optional
            full path to the file for alternative flat images
        dark_name: str, optional
            full path to the file for alternative dark images
        use_fake_flat: boolean, optional
            if fake_flat=True, a uniform flat image is generated with its value
            equal to fake_flat_val
        use_fake_dark: boolean, optional
            if fake_dark=True, a uniform dark image is generated with its value
            equal to fake_dark_val
        ds_use: boolean, optional
            if ds_use=True, the projection images is read and downsampled by
            ds_level; if ds_use=False, data of original size is read; default
            is False
        ds_level: float, optional
            downsampling level; 0.5 corresponds to a downsampling factor 2

    return: tuple of ndarray, or None
        return data, white, dark, theta if succeeds; otherwise None
    """
    if flat_name is None:
        flat_name = fn
    if dark_name is None:
        dark_name = fn

    theta = reader(fn, dtype='theta', sli=[None], cfg=cfg).astype(np.float32)
    idx = rm_redundant(theta)
    theta = theta[idx]

    if mean_axis is None:
        if ds_use:
            data = data_down_sampling(reader(fn, dtype='data',
                                             sli=[None, [sli_start, sli_end], [col_start, col_end]],
                                             cfg=cfg), [1, ds_level, ds_level]).astype(np.float32)[idx]

            if use_fake_flat:
                if fake_flat_roi is None:
                    white = data_down_sampling(fake_flat_val * np.ones([8, data.shape[1], data.shape[2]]),
                                               [1, ds_level, ds_level]).astype(np.float32)
                else:
                    white = data_down_sampling((data[:, fake_flat_roi[0]:fake_flat_roi[1],
                                                fake_flat_roi[2]:fake_flat_roi[3]].mean(axis=(1, 2), keepdims=True) *
                                                np.ones([1, data.shape[1], data.shape[2]])),
                                               [1, ds_level, ds_level]).astype(np.float32)
            else:
                white = data_down_sampling(
                    reader(flat_name, dtype='flat', sli=[None, [sli_start, sli_end], [col_start, col_end]], cfg=cfg),
                    [1, ds_level, ds_level]).astype(np.float32)

            if use_fake_dark:
                dark = data_down_sampling(fake_dark_val * np.ones([8, data.shape[1], data.shape[2]]),
                                          [1, ds_level, ds_level]).astype(np.float32)
            else:
                dark = data_down_sampling(
                    reader(dark_name, dtype='dark', sli=[None, [sli_start, sli_end], [col_start, col_end]], cfg=cfg),
                    [1, ds_level, ds_level]).astype(np.float32)
        else:
            data = reader(fn, dtype='data', sli=[None, [sli_start, sli_end], [col_start, col_end]], cfg=cfg).astype(
                np.float32)[idx]

            if use_fake_flat:
                if fake_flat_roi is None:
                    white = fake_flat_val * np.ones([8, data.shape[1], data.shape[2]], dtype=np.float32)
                else:
                    white = data[:, fake_flat_roi[0]:fake_flat_roi[1],
                                 fake_flat_roi[2]:fake_flat_roi[3]].mean(axis=(1, 2), keepdims=True) * \
                            np.ones([1, data.shape[1], data.shape[2]], dtype=np.float32)
            else:
                white = reader(flat_name, dtype='flat', sli=[None, [sli_start, sli_end], [col_start, col_end]],
                               cfg=cfg).astype(np.float32)

            if use_fake_dark:
                dark = fake_dark_val * np.ones([8, data.shape[1], data.shape[2]], dtype=np.float32)
            else:
                dark = reader(dark_name, dtype='dark', sli=[None, [sli_start, sli_end], [col_start, col_end]],
                              cfg=cfg).astype(np.float32)
    else:
        if ds_use:
            data = data_down_sampling(
                reader(fn, dtype='data', sli=[None, [sli_start, sli_end], [col_start, col_end]], cfg=cfg),
                [1, ds_level, ds_level]).astype(np.float32)[idx]
            if use_fake_flat:
                white = data_down_sampling(fake_flat_val * np.ones([data.shape[1], data.shape[2]]),
                                           [ds_level, ds_level]).astype(np.float32)
            else:
                white = data_down_sampling(
                    reader(flat_name, dtype='flat', sli=[None, [sli_start, sli_end], [col_start, col_end]], cfg=cfg),
                    [1, ds_level, ds_level]).mean(axis=0).astype(np.float32)

            if use_fake_dark:
                dark = data_down_sampling(fake_dark_val * np.ones([data.shape[1], data.shape[2]]),
                                          [ds_level, ds_level]).astype(np.float32)
            else:
                dark = data_down_sampling(
                    reader(dark_name, dtype='dark', sli=[None, [sli_start, sli_end], [col_start, col_end]], cfg=cfg),
                    [1, ds_level, ds_level]).mean(axis=0).astype(np.float32)
            data[:] = (data - dark[np.newaxis, :]) / (white[np.newaxis, :] - dark[np.newaxis, :])[:]
            data[np.isinf(data)] = 0
            data[np.isnan(data)] = 0
            data = data.mean(axis=mean_axis).astype(np.float32)
        else:
            data = reader(fn, dtype='data', sli=[None, [sli_start, sli_end], [col_start, col_end]], cfg=cfg).astype(
                np.float32)[idx]
            if use_fake_flat:
                white = fake_flat_val * np.ones([data.shape[1], data.shape[2]], dtype=np.float32)
            else:
                white = reader(flat_name, dtype='flat', sli=[None, [sli_start, sli_end], [col_start, col_end]],
                               cfg=cfg).mean(axis=0).astype(np.float32)

            if use_fake_dark:
                dark = fake_dark_val * np.ones([data.shape[1], data.shape[2]], dtype=np.float32)
            else:
                dark = reader(dark_name, dtype='dark', sli=[None, [sli_start, sli_end], [col_start, col_end]],
                              cfg=cfg).mean(axis=0).astype(np.float32)
            data[:] = (data - dark[np.newaxis, :]) / (white[np.newaxis, :] - dark[np.newaxis, :])[:]
            data[np.isinf(data)] = 0
            data[np.isnan(data)] = 0
            data = data.mean(axis=mean_axis).astype(np.float32)
    gc.collect()
    return data, white, dark, theta


def retrieve_phase(data, pixel_size=1e-4, dist=50, energy=20,
                   alpha=1e-3, pad=True, filter='paganin'):
    if filter == 'paganin':
        data[:] = tomopy.prep.phase.retrieve_phase(data, pixel_size=pixel_size,
                                                   dist=dist, energy=energy,
                                                   alpha=alpha, pad=pad)[:]
    elif filter == 'bronnikov':
        data[:] = (1 - data)[:]
        data[:] = tomopy.prep.phase.retrieve_phase(data, pixel_size=pixel_size,
                                                   dist=dist, energy=energy,
                                                   alpha=alpha, pad=pad)[:]
    return data


def rm_redundant(ang):
    dang = np.diff(ang, prepend=0)
    idx = (dang > 0.01)
    if np.argmax(idx) > 0:
        idx[np.argmax(idx) - 1] = True
    return idx


def run_engine(**kwargs):
    """
    kwargs: dictionary
        This is the reconstruction configuration dictionary tomo_recon_param_dict
        input from tomo_recon_gui
    """
    file_raw_data_top_dir = kwargs['file_params']['raw_data_top_dir']
    file_data_cen_dir = kwargs['file_params']['data_center_dir']
    file_recon_top_dir = kwargs['file_params']['recon_top_dir']
    file_debug_top_dir = kwargs['file_params']['debug_top_dir']
    file_alt_flat_fn = kwargs['file_params']['alt_flat_file']
    file_alt_dark_fn = kwargs['file_params']['alt_dark_file']
    file_wedge_ang_auto_det_ref_fn = kwargs['file_params']['wedge_ang_auto_det_ref_fn']
    file_cfg = kwargs['file_params']['io_confg']
    reader = kwargs['file_params']['reader']
    dim_info = kwargs['file_params']['info_reader']
    use_struc_h5_reader = kwargs['file_params']['use_struc_h5_reader']

    data_scan_id = kwargs["data_params"]['scan_id']
    data_ds_level = kwargs["data_params"]['downsample']
    data_rot_cen = kwargs["data_params"]['rot_cen']
    data_cen_win_s = kwargs["data_params"]['cen_win_s']
    data_cen_win_w = kwargs["data_params"]['cen_win_w']
    data_sli_s = kwargs["data_params"]['sli_s']
    data_sli_e = kwargs["data_params"]['sli_e']
    data_col_s = kwargs["data_params"]['col_s']
    data_col_e = kwargs["data_params"]['col_e']
    data_fake_flat_val = kwargs["data_params"]['fake_flat_val']
    data_fake_dark_val = kwargs["data_params"]['fake_dark_val']
    fake_flat_roi = kwargs["data_params"]["fake_flat_roi"]
    data_chunk_sz = kwargs["data_params"]['chunk_sz']
    data_margin = kwargs["data_params"]['margin']
    data_flat_blur_kernel = kwargs["data_params"]["blur_kernel"]
    data_zinger_val = kwargs["data_params"]['zinger_val']
    data_mask_ratio = kwargs["data_params"]['mask_ratio']
    data_wedge_missing_s = kwargs["data_params"]['wedge_missing_s']
    data_wedge_missing_e = kwargs["data_params"]['wedge_missing_e']
    data_wedge_col_s = kwargs["data_params"]["wedge_col_s"]
    data_wedge_col_e = kwargs["data_params"]["wedge_col_e"]
    data_wedge_ang_auto_det_thres = kwargs["data_params"]['wedge_ang_auto_det_thres']

    rec_type = kwargs['recon_config']['recon_type']
    if data_ds_level == 1:
        rec_use_ds = False
    else:
        rec_use_ds = True
    rec_use_debug = kwargs['recon_config']['use_debug']
    rec_use_alt_flat = kwargs['recon_config']['use_alt_flat']
    rec_use_alt_dark = kwargs['recon_config']['use_alt_dark']
    rec_use_fake_flat = kwargs['recon_config']['use_fake_flat']
    rec_use_fake_dark = kwargs['recon_config']['use_fake_dark']
    rec_use_blur_flat = kwargs["recon_config"]["use_flat_blur"]
    rec_use_rm_zinger = kwargs['recon_config']['use_rm_zinger']
    rec_use_mask = kwargs['recon_config']['use_mask']
    rec_use_wedge_ang_auto_det = kwargs['recon_config']['use_wedge_ang_auto_det']
    is_wedge = kwargs['recon_config']['is_wedge']

    flt_param_dict = kwargs["flt_params"]
    alg_param_dict = kwargs["alg_params"]

    if not rec_use_alt_flat:
        file_alt_flat_fn = None
    if not rec_use_alt_dark:
        file_alt_dark_fn = None

    if rec_type == 'Trial Cent':
        file_raw_fn, file_recon_template = get_file(file_raw_data_top_dir, data_scan_id, file_cfg)

        data, white, dark, theta = read_data(reader, file_raw_fn, file_cfg,
                                             sli_start=data_sli_s, sli_end=data_sli_s + 20,
                                             col_start=data_col_s, col_end=data_col_e,
                                             flat_name=file_alt_flat_fn, dark_name=file_alt_dark_fn,
                                             use_fake_flat=rec_use_fake_flat, use_fake_dark=rec_use_fake_dark,
                                             fake_flat_val=data_fake_flat_val, fake_dark_val=data_fake_dark_val,
                                             ds_use=rec_use_ds, ds_level=data_ds_level)
        theta = theta * np.pi / 180.0
        dim = data.shape

        if rec_use_rm_zinger:
            data[:] = tomopy.misc.corr.remove_outlier(data, data_zinger_val, size=15, axis=0)[:]
            white[:] = tomopy.misc.corr.remove_outlier(white, data_zinger_val, size=15, axis=0)[:]

        if rec_use_blur_flat:
            white[:] = gf(white, data_flat_blur_kernel)[:]
        data[:] = normalize(data, white, dark)[:]

        if is_wedge:
            if rec_use_wedge_ang_auto_det:
                print('wedge_ref_file: ', file_wedge_ang_auto_det_ref_fn)
                bad_angs = get_dim_ang_range_range_sli(dim_info,
                                                       reader,
                                                       file_wedge_ang_auto_det_ref_fn,
                                                       file_cfg,
                                                       sli_start=data_sli_s,
                                                       sli_end=data_sli_s + 20,
                                                       col_start=data_wedge_col_s, col_end=data_wedge_col_e,
                                                       ds_use=rec_use_ds, ds_level=data_ds_level,
                                                       thres=data_wedge_ang_auto_det_thres)
            else:
                bad_angs = np.zeros([data.shape[0], data.shape[1]], dtype=bool)
                bad_angs[data_wedge_missing_s:data_wedge_missing_e, :] = True
            data[:] = sort_wedge(data, bad_angs, 0, 20, padval=1)[:]
            print(data.shape)
            if rec_use_debug:
                save_debug(file_debug_top_dir, '1-wedge_data.tiff', data)

        if rec_use_debug:
            save_debug(file_debug_top_dir, '2-norm_data.tiff', data)

        if 0 != len(flt_param_dict.keys()):
            for idx in sorted(flt_param_dict.keys()):
                data[:] = run_filter(data, flt_param_dict[idx])[:]
                if rec_use_debug:
                    save_debug(file_debug_top_dir,
                               '3-filter_name' + str(flt_param_dict[idx].keys()) + '_filtered_data.tiff', data)

        if if_log(flt_param_dict):
            data[:] = tomopy.prep.normalize.minus_log(data)[:]
            print('doing log')
            if "remove cupping" in flt_param_dict.keys():
                params = translate_params(flt_param_dict["remove cupping"]['params'])
                data -= params['cc']
                print("running remove cupping")

        if is_wedge:
            data[:] = sort_wedge(data, bad_angs, 0, 20, padval=0)[:]

        if rec_use_debug:
            save_debug(file_debug_top_dir, '4-log_data.tiff', data)

        overwrite_dir(file_data_cen_dir)

        write_center(data[:, int(dim[1] / 2) - 1:int(dim[1] / 2) + 1, :],
                     theta, dpath=file_data_cen_dir,
                     cen_range=(data_cen_win_s, data_cen_win_s + data_cen_win_w, 0.5),
                     mask=rec_use_mask, ratio=data_mask_ratio,
                     algorithm=(alg_param_dict['algorithm']
                                if (alg_param_dict['algorithm'] != 'astra') else tomopy.astra),
                     **(translate_params(alg_param_dict['params'])))
        rec_use_logging = True
        if rec_use_logging:
            fout = os.path.join(os.path.dirname(file_raw_fn), ''.join(os.path.basename(file_raw_fn).split('.')[:-1]) + \
                                '_finding_cneter_log.txt')
            with open(fout, "w") as fo:
                for k, v in kwargs.items():
                    fo.write(str(k) + ': ' + str(v) + '\n\n')
        return 0
    else:
        state = 1
        if is_wedge:
            if rec_use_wedge_ang_auto_det:
                bad_angs = get_dim_ang_range_range_sli(dim_info, reader, file_wedge_ang_auto_det_ref_fn,
                                                       file_cfg, sli_start=0,
                                                       sli_end=None, col_start=data_wedge_col_s,
                                                       col_end=data_wedge_col_e,
                                                       ds_use=rec_use_ds, ds_level=data_ds_level,
                                                       thres=data_wedge_ang_auto_det_thres)
            else:
                bad_angs = np.zeros([data.shape[0], data.shape[1]])
                bad_angs[data_wedge_missing_s:data_wedge_missing_e, :] = 1

        file_raw_fn, file_recon_template = get_file(file_raw_data_top_dir, data_scan_id, file_cfg,
                                                    recon_top_dir=file_recon_top_dir)
        print(file_raw_fn, file_recon_template)
        dim = dim_info(file_raw_fn, dtype='data', cfg=file_cfg)

        if data_chunk_sz >= (data_sli_e - data_sli_s):
            data_chunk_sz = (data_sli_e - data_sli_s)
            num_chunk = 1
        else:
            num_chunk = np.int((data_sli_e - data_sli_s) / (data_chunk_sz - 2 * data_margin)) + 1

        for ii in range(num_chunk):
            try:
                if ii == 0:
                    sli_start = data_sli_s
                    sli_end = data_sli_s + data_chunk_sz
                else:
                    sli_start = data_sli_s + ii * (data_chunk_sz - 2 * data_margin)
                    sli_end = sli_start + data_chunk_sz
                    if sli_end > data_sli_e:
                        sli_end = data_sli_e
                    if sli_end > dim[1]:
                        sli_end = dim[1]

                if (sli_end - sli_start) <= data_margin:
                    print('skip')
                    break
                else:
                    data, white, dark, theta = read_data(reader,
                                                         file_raw_fn, file_cfg,
                                                         sli_start=sli_start, sli_end=sli_end,
                                                         col_start=data_col_s, col_end=data_col_e,
                                                         flat_name=file_alt_flat_fn, dark_name=file_alt_dark_fn,
                                                         use_fake_flat=rec_use_fake_flat,
                                                         use_fake_dark=rec_use_fake_dark,
                                                         fake_flat_val=data_fake_flat_val,
                                                         fake_dark_val=data_fake_dark_val,
                                                         fake_flat_roi=fake_flat_roi,
                                                         ds_use=rec_use_ds, ds_level=data_ds_level)

                    if is_wedge:
                        data[:] = sort_wedge(data, bad_angs, sli_start, sli_end, padval=1)[:]

                    theta = theta * np.pi / 180
                    if rec_use_rm_zinger:
                        data[:] = tomopy.misc.corr.remove_outlier(data, data_zinger_val, size=15, axis=0)[:]
                        white[:] = tomopy.misc.corr.remove_outlier(white, data_zinger_val, size=15, axis=0)[:]

                    if rec_use_blur_flat:
                        white[:] = gf(white, data_flat_blur_kernel)[:]
                    data[:] = normalize(data, white, dark)[:]

                    if 0 != len(flt_param_dict.keys()):
                        for fp_key in sorted(flt_param_dict.keys()):
                            data[:] = run_filter(data, flt_param_dict[fp_key])[:]

                    if if_log(flt_param_dict):
                        data[:] = tomopy.prep.normalize.minus_log(data)[:]
                        print('doing log')
                        if "remove cupping" in flt_param_dict.keys():
                            params = translate_params(flt_param_dict["remove cupping"]['params'])
                            data -= params['cc']
                            print("running remove cupping")

                    if is_wedge:
                        data[:] = sort_wedge(data, bad_angs, sli_start, sli_end, padval=0)[:]

                    data_recon = tomopy.recon(data, theta, center=data_rot_cen,
                                              algorithm=(alg_param_dict['algorithm']
                                                         if (alg_param_dict['algorithm'] != 'astra') else tomopy.astra),
                                              **(translate_params(alg_param_dict['params'])))
                    if rec_use_mask:
                        data_recon = tomopy.circ_mask(data_recon, 0, ratio=data_mask_ratio)

                    write_tiff_stack(data_recon[int(data_margin):(sli_end - sli_start - int(data_margin)), :, :],
                                     axis=0, fnt=file_recon_template,
                                     start=sli_start + int(data_margin),
                                     overwrite=True)
                    del (data)
                    del (white)
                    del (dark)
                    del (theta)
                    del (data_recon)
                    gc.collect()
                    print('chunk ', ii, ' reconstruction is saved')
                    print(time.asctime())
            except Exception as e:
                state = 0
                print(type(e))
                print(e.args)
        if state == 1:
            print('Reconstruction finishes!')
            rec_logging = True
            if rec_logging is True:
                fout = Path(file_recon_template).parents[0]/ \
                       (''.join(os.path.basename(file_raw_fn).split('.')[:-1]) + '_recon_log.txt')
                fo = open(fout, "w")
                for k, v in kwargs.items():
                    fo.write(str(k) + ': ' + str(v) + '\n\n')
                fo.close()
            return 0
        else:
            print('Reconstruction is terminated due to error.')
            return -1


def run_filter(data, flt):
    flt_name = flt['filter_name']
    params = translate_params(flt['params'])
    print('running', flt_name)
    if flt_name == "denoise: wiener":
        psfw = int(params['psf'])
        params['psf'] = np.ones([psfw, psfw]) / (psfw ** 2)
        for ii in range(data.shape[0]):
            data[ii] = skr.wiener(data[ii], params['psf'],
                                  params['balance'], reg=params['reg'],
                                  is_real=params['is_real'], clip=params['clip'])[:]
    elif flt_name == "denoise: median":
        data[:] = median(data, size=(int(params['size angle']),
                                     int(params['size y']),
                                     int(params['size x'])))[:]
    elif flt_name == "denoise: unsupervised_wiener":
        psfw = int(params['psf'])
        params['psf'] = np.ones([psfw, psfw]) / (psfw ** 2)
        for ii in range(data.shape[0]):
            data[ii], _ = skr.unsupervised_wiener(data[ii], params['psf'],
                                                  reg=params['reg'], user_params=params['user_params'],
                                                  is_real=params['is_real'], clip=params['clip'])[:]
    elif flt_name == "denoise: denoise_nl_means":
        for ii in range(data.shape[0]):
            data[ii] = skr.denoise_nl_means(data[ii], **params, preserve_range=None)[:]
    elif flt_name == "denoise: denoise_tv_bregman":
        for ii in range(data.shape[0]):
            data[ii] = skr.denoise_tv_bregman(data[ii], params['weight'],
                                              max_iter=params['max_iter'], eps=params['eps'],
                                              isotropic=params['isotropic'],
                                              multichannel=params['multichannel'])[:]
    elif flt_name == "denoise: denoise_tv_chambolle":
        for ii in range(data.shape[0]):
            data[ii] = skr.denoise_tv_chambolle(data[ii], params['weight'],
                                                n_iter_max=params['n_iter_max'], eps=params['eps'],
                                                multichannel=params['multichannel'])[:]
    elif flt_name == "denoise: denoise_bilateral":
        for ii in range(data.shape[0]):
            data[ii] = skr.denoise_bilateral(data[ii], **params)[:]
    elif flt_name == "denoise: denoise_wavelet":
        for ii in range(data.shape[0]):
            data[ii] = skr.denoise_wavelet(data[ii], **params)[:]
    elif flt_name == "flatting bkg":
        data[:] = tomopy.prep.normalize.normalize_bg(data, air=params['air'])[:]
    elif flt_name == "stripe_removal: vo":
        for key in params.keys():
            if key in ["la_size", "sm_size"]:
                params[key] = int(params[key])
        data[:] = tomopy.prep.stripe.remove_all_stripe(data, **params)[:]
    elif flt_name == "stripe_removal: ti":
        data[:] = tomopy.prep.stripe.remove_stripe_ti(data, **params)[:]
    elif flt_name == "stripe_removal: sf":
        params['size'] = int(params['size'])
        data[:] = tomopy.prep.stripe.remove_stripe_sf(data, **params)[:]
    elif flt_name == "stripe_removal: fw":
        params['level'] = int(params['level'])
        data[:] = tomopy.prep.stripe.remove_stripe_fw(data, **params)[:]
    elif flt_name == "phase retrieval":
        data[:] = retrieve_phase(data, **params)[:]
    return data


def save_debug(debug_dir, debug_fn, data):
    if not os.path.exists(debug_dir):
        os.makedirs(debug_dir)
    tifffile.imsave(os.path.join(debug_dir, debug_fn), data.astype(np.float32))


def sort_wedge(data, bad_angs, sli_start, sli_end, padval=0):
    bad_angs = match_size(data, bad_angs)
    for ii in range(sli_start, sli_end):
        data[bad_angs[:, ii], ii - sli_start, :] = padval
    return data


def translate_params(params):
    for key, param in params.items():
        if param == 'None':
            params[key] = None
        elif param == 'True':
            params[key] = True
        elif param == 'False':
            params[key] = False
    return params


def match_size(data, bad_angs):
    if bad_angs.shape[0] < data.shape[0]:
        bad_angs = np.pad(bad_angs, [[0, data.shape[0] - bad_angs.shape[0]],
                                     [0, 0]], mode='edge')
    elif bad_angs.shape[0] > data.shape[0]:
        bad_angs = bad_angs[:data.shape[0], ...]
    return bad_angs


def overwrite_dir(path):
    if os.path.isdir(path):
        if os.path.exists(path):
            shutil.rmtree(path)
            os.makedirs(path)
        else:
            os.makedirs(path)
        return 0
    else:
        return -1


def write_center(tomo, theta, dpath=Path('tmp/center'), cen_range=None, ind=None,
                 mask=False, ratio=1., sinogram_order=False, algorithm='gridrec',
                 filter_name='parzen', **kwargs):
    """
    Save images reconstructed with a range of rotation centers.

    Helps finding the rotation center manually by visual inspection of
    images reconstructed with a set of different centers.The output
    images are put into a specified folder and are named by the
    center position corresponding to the image.

    Parameters
    ----------
    tomo : ndarray
        3D tomographic data.
    theta : array
        Projection angles in radian.
    dpath : str, optional
        Folder name to save output images.
    cen_range : list, optional
        [start, end, step] Range of center values.
    ind : int, optional
        Index of the slice to be used for reconstruction.
    mask : bool, optional
        If ``True``, apply a circular mask to the reconstructed image to
        limit the analysis into a circular region.
    ratio : float, optional
        The ratio of the radius of the circular mask to the edge of the
        reconstructed image.
    sinogram_order: bool, optional
        Determins whether data is a stack of sinograms (True, y-axis first axis)
        or a stack of radiographs (False, theta first axis).
    algorithm : {str, function}
        One of the following string values.

        'art'
            Algebraic reconstruction technique :cite:`Kak:98`.
        'bart'
            Block algebraic reconstruction technique.
        'fbp'
            Filtered back-projection algorithm.
        'gridrec'
            Fourier grid reconstruction algorithm :cite:`Dowd:99`,
            :cite:`Rivers:06`.
        'mlem'
            Maximum-likelihood expectation maximization algorithm
            :cite:`Dempster:77`.
        'osem'
            Ordered-subset expectation maximization algorithm
            :cite:`Hudson:94`.
        'ospml_hybrid'
            Ordered-subset penalized maximum likelihood algorithm with
            weighted linear and quadratic penalties.
        'ospml_quad'
            Ordered-subset penalized maximum likelihood algorithm with
            quadratic penalties.
        'pml_hybrid'
            Penalized maximum likelihood algorithm with weighted linear
            and quadratic penalties :cite:`Chang:04`.
        'pml_quad'
            Penalized maximum likelihood algorithm with quadratic penalty.
        'sirt'
            Simultaneous algebraic reconstruction technique.
        'tv'
            Total Variation reconstruction technique
            :cite:`Chambolle:11`.
        'grad'
            Gradient descent method with a constant step size
        'tikh'
            Tikhonov regularization with identity Tikhonov matrix.            


    filter_name : str, optional
        Name of the filter for analytic reconstruction.

        'none'
            No filter.
        'shepp'
            Shepp-Logan filter (default).
        'cosine'
            Cosine filter.
        'hann'
            Cosine filter.
        'hamming'
            Hamming filter.
        'ramlak'
            Ram-Lak filter.
        'parzen'
            Parzen filter.
        'butterworth'
            Butterworth filter.
        'custom'
            A numpy array of size `next_power_of_2(num_detector_columns)/2`
            specifying a custom filter in Fourier domain. The first element
            of the filter should be the zero-frequency component.
        'custom2d'
            A numpy array of size `num_projections*next_power_of_2(num_detector_columns)/2`
            specifying a custom angle-dependent filter in Fourier domain. The first element
            of each filter should be the zero-frequency component.
    """
    tomo = tomopy.util.dtype.as_float32(tomo)
    theta = tomopy.util.dtype.as_float32(theta)

    if sinogram_order:
        dy, dt, dx = tomo.shape
    else:
        dt, dy, dx = tomo.shape
    if ind is None:
        ind = dy // 2
    if cen_range is None:
        center = np.arange(dx / 2 - 5, dx / 2 + 5, 0.5)
    else:
        center = np.arange(*cen_range)

    stack = tomopy.util.dtype.empty_shared_array((len(center), dt, dx))

    for m in range(center.size):
        if sinogram_order:
            stack[m] = tomo[ind]
        else:
            stack[m] = tomo[:, ind, :]

    # Reconstruct the same slice with a range of centers.
    rec = tomopy.recon(stack,
                       theta,
                       center=center,
                       sinogram_order=True,
                       algorithm=algorithm,
                       nchunk=1, **kwargs)

    # Apply circular mask.
    if mask is True:
        rec = tomopy.circ_mask(rec, axis=0)

    # Save images to a temporary folder.
    dpath = os.path.abspath(dpath)
    if not os.path.exists(dpath):
        os.makedirs(dpath)

    for m in range(len(center)):
        tomopy.util.misc.write_tiff(data=rec[m], fname=dpath, digit='{0:.2f}'.format(center[m]))


def write_tiff_stack(img_stack, axis=0, fnt=None, start=0, overwrite=True):
    if axis == 0:
        for ii in range(img_stack.shape[axis]):
            tifffile.imsave(fnt.format(str(start + ii).zfill(5)), img_stack[ii].astype(np.float32))
    elif axis == 1:
        for ii in range(img_stack.shape[axis]):
            tifffile.imsave(fnt.format(str(start + ii).zfill(5)), img_stack[:, ii, :].astype(np.float32))
    elif axis == 2:
        for ii in range(img_stack.shape[axis]):
            tifffile.imsave(fnt.format(str(start + ii).zfill(5)), img_stack[:, :, ii].astype(np.float32))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Aug 23 23:27:42 2020

@author: xiao
"""

IO_TOMO_CFG_DEFAULT = {'structured_h5_reader':{
                       'io_data_structure':{'data_path':'/img_tomo',
                                            'flat_path':'/img_bkg',
                                            'dark_path':'/img_dark',
                                            'theta_path':'/angle'},
                       'io_data_info':{'item00_path':'/img_tomo',
                                       'item01_path':'/angle',
                                       'item02_path':'/Magnification',
                                       'item03_path':'/Pixel Size',
                                       'item04_path':'/X_eng',
                                       'item05_path':'/note',
                                       'item06_path':'/scan_time',
                                       'item07_path':''}
                                               },
                       'customized_reader':{
                          'user_tomo_reader':''   
                                              },
                       'tomo_raw_fn_template':'fly_scan_id_{}.h5'
                      }


IO_XANES2D_CFG_DEFAULT = {'structured_h5_reader':{
                          'io_data_structure':{'data_path':'/img_xanes',
                                               'flat_path':'/img_bkg',
                                               'dark_path':'/img_dark',
                                               'eng_path':'/X_eng'},
                          'io_data_info':{'item00_path':'/img_xanes',
                                          'item01_path':'/Magnification',
                                          'item02_path':'/Pixel Size',
                                          'item03_path':'/X_eng',
                                          'item04_path':'/note',
                                          'item05_path':'/scan_time',
                                          'item06_path':'',
                                          'item07_path':''}
                                                  },
                          'customized_reader':{
                          'user_xanes2D_reader':''    
                                              },
                          'xanes2D_raw_fn_template':'xanes_scan2_id_{}.h5'
                         }


IO_XANES3D_CFG_DEFAULT = {'structured_h5_reader':{
                          'io_data_structure':{'data_path':'/img_tomo',
                                               'flat_path':'/img_bkg',
                                               'dark_path':'/img_dark',
                                               'eng_path':'/X_eng'},
                          'io_data_info':{'item00_path':'/img_tomo',
                                          'item01_path':'/angle',
                                          'item02_path':'/Magnification',
                                          'item03_path':'/Pixel Size',
                                          'item04_path':'/X_eng',
                                          'item05_path':'/note',
                                          'item06_path':'/scan_time',
                                          'item07_path':''}
                                                  },
                          'customized_reader':{
                          'user_xanes3D_reader':''    
                                              },
                          'tomo_raw_fn_template':'fly_scan_id_{}.h5',
                          'xanes3D_recon_dir_template':'recon_fly_scan_id_{}',
                          'xanes3D_recon_fn_template':'recon_fly_scan_id_{0}_{1}.tiff'
                         }


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Sep  7 12:46:23 2019

@author: xiao
"""

from pathlib import Path
import numpy as np
import pandas as pd
from scipy.ndimage import median_filter, uniform_filter
from scipy.interpolate import splrep, splev
import h5py, os, time
from copy import deepcopy
import multiprocess as mp
from tomopy.util.mproc import distribute_jobs

from .misc import msgit
from . import xanes_math as xm

"""
    This class include xanes spectrum analyses. These analyses are based on
    filtered xanes spectra. The basic functions include:
        1. curve fitting
        2. peak fitting
    The analyses in this package include:

        1. Subtracting pre_edge backgroud
        2. Normalizing spectrum
        3. Finding e0
        4. Calculating edge jump
        5. Calculating pre_edge and post_edge statistics
        6. Finding peak (whiteline)
        7, PCA analysis
        8. Linear Combinatiotn analysis
"""


class xanes_analysis():
    def __init__(self, spectrum, eng, preset_edge_eng,
                 pre_es=None, pre_ee=-50,
                 post_es=100, post_ee=None,
                 edge_jump_threshold=5,
                 pre_edge_threshold=1):
        self.spec = spectrum
        self.eng = eng
        self.preset_edge_eng = preset_edge_eng
        self.model = {'edge': {}, 'wl': {}}
        self.lcf_use = False
        self.lcf_constr_use = True
        self.lcf_ref_spec = None
        self.lcf_ref = None
        self.lcf_fit = None
        self.lcf_model = {'constr': self.lcf_constr_use, 'ref': self.lcf_ref, 'rlt': None}

        if pre_es is None:
            pre_es = eng[0] - preset_edge_eng
        self.pre_es = preset_edge_eng + pre_es
        self.pre_ee = preset_edge_eng + pre_ee
        self.pre_es_idx = xm.index_of(self.eng, self.pre_es)
        self.pre_ee_idx = xm.index_of(self.eng, self.pre_ee)

        self.post_es = preset_edge_eng + post_es
        if post_ee is None:
            post_ee = eng[-1] - preset_edge_eng
        self.post_ee = preset_edge_eng + post_ee
        self.post_es_idx = xm.index_of(self.eng, self.post_es)
        self.post_ee_idx = xm.index_of(self.eng, self.post_ee)

        self.edge_jump_thres = edge_jump_threshold
        self.fitted_edge_thres = pre_edge_threshold
        self.auto_e0 = [None]
        self.pre_edge_fit = None
        self.post_edge_fit = None
        self.edge_jump_mask = None
        self.fitted_edge_mask = None
        self.norm_spec = None
        self.fitted_edge_flted_spec = None
        self.edge_jump_flted_spec = None
        self.removed_img_ids = []
        self.post_edge_fit_rlt = np.empty([])
        self.pre_edge_fit_rlt = np.empty([])
        self.pre_edge_sd_map = np.empty([])
        self.post_edge_sd_map = np.empty([])
        self.pre_edge_mean_map = np.empty([])
        self.post_edge_mean_map = np.empty([])
        self.pre_avg = np.empty([])
        self.post_avg = np.empty([])
        self.edge_jump_map = np.empty([])
        self.centroid_of_eng = np.empty([])
        self.centroid_of_eng_rel_wl = np.empty([])
        self.weighted_atten = np.empty([])
        self.weighted_eng = np.empty([])
        self.edge0p5_pos_fit = np.empty([])
        self.edge_fit_rlt = np.empty([])
        self.wl_fit_rlt = np.empty([])

        self.wl_pos_dir = np.empty([])
        self.direct_wl_ph = np.empty([])
        self.wl_pos_fit = np.empty([])
        self.edge_pos = np.empty([])

        self.pre_edge_fitted = False
        self.post_edge_fitted = False
        self.spec_normalized = False
        self.spec_linear_fitted = False
        self.spec_pca_fitted = False

    def apply_edge_jump_filter(self):
        """
        return: ndarray, same dimension as spectrum.shape
        """
        return self.spec * self.edge_jump_mask

    def apply_external_mask(self):
        return self.spec * self.ext_mask

    def apply_fitted_edge_filter(self):
        """
        return: ndarray, same dimension as spectrum.shape
        """
        return self.spec * self.fitted_edge_mask

    def cal_pre_edge_sd(self):
        """
        return: ndarray, pre_edge_sd has dimension of spectrum.shape[1:]
        """
        self.pre_edge_sd_map = self.spec[self.pre_es_idx:self.pre_ee_idx,
                               :].std(axis=0)

    def cal_post_edge_sd(self):
        """
        return: ndarray, post_edge_sd has dimension of spectrum.shape[1:]
        """
        self.post_edge_sd_map = self.spec[self.post_es_idx:self.post_ee_idx,
                                :].std(axis=0)

    def cal_pre_edge_mean(self):
        """
        return: ndarray, pre_edge_sd has dimension of spectrum.shape[1:]
        """
        self.pre_edge_mean_map = self.spec[self.pre_es_idx:self.pre_ee_idx, :].mean(axis=0)

    def cal_post_edge_mean(self):
        """
        return: ndarray, post_edge_sd has dimension of spectrum.shape[1:]
        """
        self.post_edge_mean_map = self.spec[self.post_es_idx:self.post_ee_idx,
                                  :].mean(axis=0)

    def cal_fit(self, fit_coef, eng=None, reshape=True):
        if eng is None:
            eng = self.eng
        if not isinstance(eng, np.ndarray):
            eng = np.array(eng)
        fit = xm.eval_polynd(fit_coef, eng)
        if reshape:
            if eng.shape:
                return fit.reshape([eng.shape[0], *self.spec.shape[1:]])
            else:
                return fit.reshape([*self.spec.shape[1:]])
        else:
            return fit

    @msgit(wd=100, fill='+')
    def cal_pre_edge_fit(self, eng=None, reshape=True):
        """
        inputs:
            eng: 1D array-like; full energy list of the spectrum
            pre_edge_fit_coef: array-like; pixel-wise pre_edge fit coef
        ouputs:
            pre_edge_fit: array-like; pixel-wise line profile over eng
        """
        self.fit_pre_edge()
        return self.cal_fit(self.pre_edge_fit_rlt[0], eng=eng, reshape=reshape)

    @msgit(wd=100, fill='+')
    def cal_post_edge_fit(self, eng=None, reshape=True):
        """
        inputs:
            eng: 1D array-like; full energy list of the spectrum
            post_edge_fit_coef: array-like; pixel-wise post_edge fit coef
        ouputs:
            post_edge_fit: array-like; pixel-wise line profile over eng
        """
        self.fit_post_edge()
        return self.cal_fit(self.post_edge_fit_rlt[0], eng=eng, reshape=reshape)

    @msgit(wd=100, fill='+')
    def cal_edge_jump_map(self):
        """
        return: ndarray, edge_jump_map has dimension of spectrum.shape[1:]
        """
        self.edge_jump_map = (self.cal_post_edge_fit(np.array([self.preset_edge_eng])) -
                              self.cal_pre_edge_fit(np.array([self.preset_edge_eng])))

    def cal_drt_wl_peak_hgt(self, peak_es, peak_ee):
        self.wl_es_idx = xm.index_of(self.eng, peak_es)
        self.wl_ee_idx = xm.index_of(self.eng, peak_ee)

        eng = self.eng[self.wl_es_idx:self.wl_ee_idx]
        for ii in range(1, len(self.spec.shape)):
            eng = eng[:, np.newaxis]
        self.direct_wl_ph = np.squeeze(np.max(self.spec[self.wl_es_idx:self.wl_ee_idx, :], axis=0))

    def cal_wgt_eng(self, eng_s):
        if not self.direct_wl_ph.shape:
            print('Please calculate the whiteline peak height first. Quit!')
            self.centroid_of_eng = np.empty([0])
            self.centroid_of_eng_rel_wl = np.empty([0])
            self.weighted_atten = np.empty([0])
            self.weighted_eng = np.empty([0])
            return
        if (not self.wl_pos_dir.shape) and (not self.wl_pos_fit.shape):
            print('Please calculate the whiteline energy first. Quit!')
            self.centroid_of_eng = np.empty([0])
            self.centroid_of_eng_rel_wl = np.empty([0])
            self.weighted_atten = np.empty([0])
            self.weighted_eng = np.empty([0])
            return
        if self.wl_pos_fit.shape:
            eng_e = self.wl_pos_fit
        else:
            eng_e = self.wl_pos_dir
        eng = deepcopy(self.eng)
        for ii in range(0, len(eng_e.shape)):
            eng = eng[:, np.newaxis]

        a = (eng >= eng_s) & (eng <= eng_e + 0.0002)
        b = (np.where(a, self.spec, 0) * eng * (np.roll(eng, -1, axis=0) - eng)).sum(axis=0)
        c = (np.where(a, self.spec, 0) * (np.roll(eng, -1, axis=0) - eng)).sum(axis=0)
        d = np.where(a, eng * (np.roll(eng, -1, axis=0) - eng), 0).sum(axis=0)
        e = (np.where(a, self.spec, 0) * np.abs(eng - eng_e) * (np.roll(eng, -1, axis=0) - eng)).sum(axis=0)

        self.centroid_of_eng = b / c
        self.centroid_of_eng_rel_wl = e / c
        self.weighted_atten = b / d
        self.weighted_eng = b / self.direct_wl_ph

    @msgit(wd=100, fill='+')
    def create_edge_jump_filter(self, eng0=None):
        """
        return: ndarray, same dimension as spectrum.shape
        """
        if eng0 is None:
            eng0 = self.edge_eng
        self.edge_jump_mask = ((self.cal_post_edge_fit(eng0) -
                                self.cal_pre_edge_fit(eng0))
                               > self.edge_jump_thres * self.pre_edge_sd_map).astype(np.int8)

    @msgit(wd=100, fill='+')
    def create_fitted_edge_filter(self):
        """
        return: ndarray, same dimension as spectrum.shape
        """
        self.fitted_edge_mask = np.any((self.cal_post_edge_fit(self.eng) -
                                        self.cal_pre_edge_fit(self.eng))
                                       > self.fitted_edge_thres * self.pre_edge_sd_map, axis=0).astype(np.int8)

    @msgit(wd=100, fill='$')
    def find_edge_0p5(self, es, ee, optimizer='model', ufac=20):
        idx_s = xm.index_of(self.eng, es)
        idx_e = xm.index_of(self.eng, ee)
        us_eng = np.linspace(self.eng[idx_s], self.eng[idx_e],
                             num=(idx_e - idx_s + 1) * ufac) - self.model['edge']['eoff']
        if optimizer == 'model':
            self.edge0p5_pos_fit = xm.find_fit_val(self.model['edge']['model'],
                                                   np.array(self.model['edge']['fit_rlt'][0]),
                                                   us_eng, v=0.5).reshape(self.spec.shape[1:]) + \
                                   self.model['edge']['eoff']
        elif optimizer == 'direct':
            self.edge0p5_pos_dir = xm.lookup(self.eng[idx_s:idx_e],
                                             self.norm_spec[idx_s:idx_e].reshape(idx_e - idx_s, -1), 0.5). \
                reshape(self.spec.shape[1:])

    @msgit(wd=100, fill='$')
    def find_edge_50(self, ees, eee, pes, pee, optimizer='model', ufac=20):
        idx_es = xm.index_of(self.eng, ees)
        idx_ee = xm.index_of(self.eng, eee)
        idx_ps = xm.index_of(self.eng, pes)
        idx_pe = xm.index_of(self.eng, pee)
        us_e_eng = np.linspace(self.eng[idx_es], self.eng[idx_ee], num=(idx_ee - idx_es) * ufac + 1)
        us_p_eng = np.linspace(self.eng[idx_ps], self.eng[idx_pe], num=(idx_pe - idx_ps) * ufac + 1)
        if optimizer == 'both':
            print('both')
            self.edge50_pos_fit = xm.find_50_peak(self.model['edge']['model'], us_e_eng-self.model['edge']['eoff'],
                                                  np.array(self.model['edge']['fit_rlt'][0]),
                                                  self.model['wl']['model'], us_p_eng-self.model['wl']['eoff'],
                                                  np.array(self.model['wl']['fit_rlt'][0]), ftype=optimizer). \
                    reshape(self.spec.shape[1:]) + self.model['edge']['eoff']
        elif optimizer == 'wl':
            print('wl')
            self.edge50_pos_fit = xm.find_50_peak(us_e_eng, self.eng[idx_es:idx_ee+1],
                                                  self.norm_spec[idx_es:idx_ee+1].reshape([idx_ee-idx_es+1, -1]),
                                                  self.model['wl']['model'], us_p_eng-self.model['wl']['eoff'],
                                                  np.array(self.model['wl']['fit_rlt'][0]), ftype=optimizer). \
                    reshape(self.spec.shape[1:])
        elif optimizer == 'edge':
            print('edge')
            self.edge50_pos_fit = xm.find_50_peak(self.model['edge']['model'], us_e_eng-self.model['edge']['eoff'],
                                                  np.array(self.model['edge']['fit_rlt'][0]),
                                                  us_p_eng, self.eng[idx_ps:idx_pe+1],
                                                  self.norm_spec[idx_ps:idx_pe+1].reshape([idx_pe-idx_ps+1, -1]),
                                                  ftype=optimizer). \
                    reshape(self.spec.shape[1:]) + self.model['edge']['eoff']
        elif optimizer == 'none':
            print('none')
            self.edge50_pos_fit_none = xm.find_50_peak(us_e_eng, self.eng[idx_es:idx_ee+1],
                                                  self.norm_spec[idx_es:idx_ee+1].reshape([idx_ee-idx_es+1, -1]),
                                                  us_p_eng, self.eng[idx_ps:idx_pe + 1],
                                                  self.norm_spec[idx_ps:idx_pe+1].reshape([idx_pe-idx_ps+1, -1]),
                                                  ftype=optimizer). \
                    reshape(self.spec.shape[1:])
        elif optimizer == 'direct':
            print('dir')
            self.edge50_pos_dir = xm.lookup(self.eng[idx_es:idx_ee],
                                            self.norm_spec[idx_es:idx_ee].reshape([idx_ee-idx_es, -1]),
                                            0.5*self.norm_spec[idx_ps:idx_pe].reshape([idx_pe-idx_ps, -1]).max(axis=0)). \
                    reshape(self.spec.shape[1:])

    @msgit(wd=100, fill='$')
    def find_wl(self, es, ee, optimizer='model', ufac=20):
        idx_s = xm.index_of(self.eng, es)
        idx_e = xm.index_of(self.eng, ee)
        if optimizer in 'model':
            us_eng = np.linspace(self.eng[idx_s], self.eng[idx_e],
                                 num=(idx_e - idx_s) * ufac) - self.model['wl']['eoff']
            self.wl_pos_fit = xm.find_fit_peak(self.model['wl']['model'],
                                               np.array(self.model['wl']['fit_rlt'][0]), us_eng) \
                                  .reshape(self.spec.shape[1:]) + self.model['wl']['eoff']
        elif optimizer == 'direct':
            if self.norm_spec is None:
                self.wl_pos_dir = xm.lookup(self.eng[idx_s:idx_e],
                                            self.spec[idx_s:idx_e].reshape(idx_e - idx_s, -1),
                                            self.spec[idx_s:idx_e].reshape(idx_e - idx_s, -1).max(axis=0)) \
                    .reshape(self.spec.shape[1:])
            else:
                self.wl_pos_dir = xm.lookup(self.eng[idx_s:idx_e],
                                            self.norm_spec[idx_s:idx_e].reshape(idx_e - idx_s, -1),
                                            self.norm_spec[idx_s:idx_e].reshape(idx_e - idx_s, -1).max(axis=0)) \
                    .reshape(self.spec.shape[1:])

    @msgit(wd=100, fill='$')
    def find_edge_deriv(self, es, ee, optimizer='model', ufac=20):
        idx_s = xm.index_of(self.eng, es)
        idx_e = xm.index_of(self.eng, ee)
        us_eng = np.linspace(self.eng[idx_s], self.eng[idx_e],
                             num=(idx_e - idx_s) * ufac) - self.model['edge']['eoff']
        if optimizer == 'model':
            self.edge_pos_fit = xm.find_deriv_peak(self.model['edge']['model'],
                                                   np.array(self.model['edge']['fit_rlt'][0]), us_eng) \
                                    .reshape(self.spec.shape[1:]) + self.model['edge']['eoff']
        elif optimizer == 'direct':
            if self.norm_spec is None:
                self.edge_pos_dir = self.spec[idx_s:idx_e].max(axis=0)
            else:
                self.edge_pos_dir = self.norm_spec[idx_s:idx_e].max(axis=0)

    def _finde0(energy, mu):
        if len(energy.shape) > 1:
            energy = energy.squeeze()
        if len(mu.shape) > 1:
            mu = mu.squeeze()
        dmu = np.gradient(mu) / np.gradient(energy)
        # find points of high derivative
        dmu[np.where(~np.isfinite(dmu))] = -1.0
        nmin = max(3, int(len(dmu) * 0.05))
        maxdmu = max(dmu[nmin:-nmin])
        high_deriv_pts = np.where(dmu > maxdmu * 0.1)[0]
        idmu_max, dmu_max = 0, 0
        for i in high_deriv_pts:
            if i < nmin or i > len(energy) - nmin:
                continue
            if (dmu[i] > dmu_max and
                    (i + 1 in high_deriv_pts) and
                    (i - 1 in high_deriv_pts)):
                idmu_max, dmu_max = i, dmu[i]
        return energy[idmu_max]

    def find_edge(self):
        """
        ### adopted from larch
        calculate :math:`E_0`, the energy threshold of absorption, or
        'edge energy', given :math:`\mu(E)`.

        :math:`E_0` is found as the point with maximum derivative with
        some checks to avoid spurious glitches.

        Arguments:
            energy (ndarray or group): array of x-ray energies, in eV, or group
            mu     (ndaarray or None): array of mu(E) values
            group  (group or None):    output group
            _larch (larch instance or None):  current larch session.

        Returns:
            float: Value of e0. If a group is provided, group.e0 will also be set.

        Notes:
            1. Supports :ref:`First Argument Group` convention, requiring group members `energy` and `mu`
            2. Supports :ref:`Set XAFS Group` convention within Larch or if `_larch` is set.
        """
        ids = xm.index_of(self.eng, self.pre_ee)
        ide = xm.index_of(self.eng, self.post_es)
        eng = self.eng[ids:ide]
        dim = self.spec[ids:ide, ...].shape
        spec = self.spec[ids:ide, ...].reshape([dim[0], -1])
        n_cpu = os.cpu_count()
        with mp.Pool(n_cpu - 1) as pool:
            rlt = pool.starmap(_finde0, [(eng, spec[:, ii]) for ii in np.int32(np.arange(spec.shape[1]))])
        pool.close()
        pool.join()
        self.edge_pos_dir = np.array(rlt).reshape(dim[1:])

    def find_edge_t(self):
        """calculate :math:`E_0`, the energy threshold of absorption, or
        'edge energy', given :math:`\mu(E)`.

        :math:`E_0` is found as the point with maximum derivative with
        some checks to avoid spurious glitches; using tomopy distribute_jobs
        for multiprocessing

        Arguments:
            energy (ndarray or group): array of x-ray energies, in eV, or group
            mu     (ndaarray or None): array of mu(E) values
            group  (group or None):    output group
            _larch (larch instance or None):  current larch session.

        Returns:
            float: Value of e0. If a group is provided, group.e0 will also be set.

        Notes:
            1. Supports :ref:`First Argument Group` convention, requiring group members `energy` and `mu`
            2. Supports :ref:`Set XAFS Group` convention within Larch or if `_larch` is set.
        """
        ids = xm.index_of(self.eng, self.pre_ee)
        ide = xm.index_of(self.eng, self.post_es)
        eng = self.eng[ids:ide]

        self.edge_pos_dir = np.ndarray(self.spec.shape[1:])
        distribute_jobs(self.spec[ids:ide, ...], _finde0, 0, kwargs={'energy': eng}, out=self.edge_pos_dir)

    def fit_pre_edge(self):
        """
        return: ndarray, pre_edge_fit has dimension of [2].append(list(spectrum.shape[1:]))
        """
        try:
            kernel = 5 * np.ones([len(self.spec.shape)])
            kernel[0] = 1
            self.pre_edge_fit_rlt = xm.fit_curv_polynd(self.eng[self.pre_es_idx:self.pre_ee_idx],
                                                       uniform_filter(self.spec[self.pre_es_idx:self.pre_ee_idx],
                                                                     size=kernel), 1)
            self.pre_edge_fitted = True
        except Exception as e:
            self.pre_edge_fitted = False
            print(str(e))
            print("pre_edge fitting went wrong")

    def fit_post_edge(self):
        """
        return: ndarray, post_edge_fit has dimension of [2].append(list(spectrum.shape[1:]))
        """
        try:
            kernel = 5*np.ones([len(self.spec.shape)])
            kernel[0] = 1
            self.post_edge_fit_rlt = xm.fit_curv_polynd(self.eng[self.post_es_idx:self.post_ee_idx],
                                                        uniform_filter(self.spec[self.post_es_idx:self.post_ee_idx],
                                                                      size=kernel), 1)
            self.post_edge_fitted = True
        except Exception as e:
            print(type(e))
            print(e.args)
            print(e)
            self.post_edge_fitted = False
            print("post_edge fitting went wrong")

    def fit_spec(self, es, ee, eoff, optimizer, flt_spec, on, *args, **kwargs):
        """
        Provide curve fitting api to that are based on numpy.ployfit and
        scipy.optimize.lsq.
        Parameters
        ----------
        es : flot
            energy starting point in eV.
        ee : flot
            energy end point in eV.
        eoff: float
            energy value offset in eV; for 'numpy' option, a meaningful value, e.g.,
            edge energy, is necessary to obtain good results
        optimizer : str
            'scipy': lsq based
            'numpy': numpy.polyfit based.
        flt_spec: boolean
            if median filtering the spectrum before fitting,
        on: str
            fitting on either 'raw' or 'norm'(alized) spectrum data
        *args: positional arguments to fitting functions
            'scipy':
                model: model function name
                fvars: model function's initial argument values
            'numpy':
                order: order of the polynomial fitting function
        **kwargs : keyword arguments to fitting functions
            'scipy': least_squares arguments
                bnds=None,
                ftol=1e-7,
                xtol=1e-7,
                gtol=1e-7,
                jac='3-point',
                method='trf'
            'numpy': NA

        Returns
        -------
        rlt: ndarray of object
            rlt[0]: coef of fitting curve.
            rlt[1:4]: residuals, rank, singular_values, rcond in numpy.polyfit case
            rlt[1:3]: cost, status, success in scipy.optimize.lsq case

        """
        idx_s = xm.index_of(self.eng, es)
        idx_e = xm.index_of(self.eng, ee)
        if flt_spec or (flt_spec == 'True'):
            flt_kernal = np.ones(len(self.spec.shape), dtype=np.int32)
            flt_kernal[0] = 3

            if on == "raw":
                if optimizer == "scipy":
                    fit_rlt = xm.fit_curv_scipy(self.eng[idx_s:idx_e] - eoff,
                                                median_filter(self.spec[idx_s:idx_e, :],
                                                              size=flt_kernal),
                                                *args, **kwargs)
                elif optimizer == "numpy":
                    fit_rlt = xm.fit_curv_polynd(self.eng[idx_s:idx_e] - eoff,
                                                 median_filter(self.spec[idx_s:idx_e, :],
                                                               size=flt_kernal),
                                                 *args)
            elif on == "norm":
                if optimizer == "scipy":
                    fit_rlt = xm.fit_curv_scipy(self.eng[idx_s:idx_e] - eoff,
                                                median_filter(self.norm_spec[idx_s:idx_e, :],
                                                              size=flt_kernal),
                                                *args, **kwargs)
                elif optimizer == "numpy":
                    fit_rlt = xm.fit_curv_polynd(self.eng[idx_s:idx_e] - eoff,
                                                 median_filter(self.norm_spec[idx_s:idx_e, :],
                                                               size=flt_kernal),
                                                 *args)
        else:
            if on == "raw":
                if optimizer == "scipy":
                    fit_rlt = xm.fit_curv_scipy(self.eng[idx_s:idx_e] - eoff,
                                                self.spec[idx_s:idx_e, :],
                                                *args, **kwargs)
                elif optimizer == "numpy":
                    fit_rlt = xm.fit_curv_polynd(self.eng[idx_s:idx_e] - eoff,
                                                 self.spec[idx_s:idx_e, :],
                                                 *args)
            elif on == "norm":
                if optimizer == "scipy":
                    fit_rlt = xm.fit_curv_scipy(self.eng[idx_s:idx_e] - eoff,
                                                self.norm_spec[idx_s:idx_e, :],
                                                *args, **kwargs)
                elif optimizer == "numpy":
                    fit_rlt = xm.fit_curv_polynd(self.eng[idx_s:idx_e] - eoff,
                                                 self.norm_spec[idx_s:idx_e, :],
                                                 *args)
        return fit_rlt

    def fit_edge(self, es, ee, eoff=None, optimizer='numpy',
                 flt_spec=False, on='norm', ftype='edge', order=3,
                 model='lorentzian', fvars=None,
                 bnds=None, ftol=1e-7, xtol=1e-7,
                 gtol=1e-7, jac='3-point',
                 method='trf'):
        if eoff is None:
            eoff = self.preset_edge_eng
        if optimizer == "numpy":
            self.model[ftype]['model'] = 'polynd'
            args = [order]
            kwargs = {}
        elif optimizer == "scipy":
            self.model[ftype]['model'] = model
            args = [model, fvars]
            kwargs = {'bnds': bnds, 'ftol': ftol, 'xtol': xtol, 'gtol': gtol,
                      'jac': jac, 'method': method}
        self.model[ftype]['eoff'] = eoff
        self.model[ftype]['fit_rlt'] = self.fit_spec(es, ee, eoff, optimizer, flt_spec, on,
                                                     *args, **kwargs)

    @msgit(wd=100, fill='$')
    def full_spec_preprocess(self, eng0, order=1, save_pre_post=False):
        e0_idx = xm.index_of(self.eng, eng0)
        pre = self.cal_pre_edge_fit(self.eng)
        post = self.cal_post_edge_fit(self.eng)

        self.edge_jump_mask = np.squeeze((post[e0_idx] - pre[e0_idx])
                                         > self.edge_jump_thres * self.pre_edge_sd_map).astype(np.int8)
        self.fitted_edge_mask = np.any((post - pre) > self.fitted_edge_thres * self.pre_edge_sd_map, axis=0).astype(
            np.int8)

        self.norm_spec = (self.spec - pre) / (self.cal_post_edge_fit(eng0) - self.cal_pre_edge_fit(eng0))
        self.norm_spec[np.isnan(self.norm_spec)] = 0
        self.norm_spec[np.isinf(self.norm_spec)] = 0

    def interp_ref_spec(self):
        self.lcf_ref_spec = np.ndarray([self.eng.shape[0], self.lcf_ref.index.size])
        for ii in range(self.lcf_ref.index.size):
            c = splrep(self.lcf_ref.iloc[ii]['eng'], self.lcf_ref.iloc[ii]['mu'])
            self.lcf_ref_spec[:, ii] = splev(self.eng, c)

    @msgit(wd=100, fill='+')
    def lcf(self):
        self.lcf_model['constr'] = self.lcf_constr_use
        self.lcf_model['ref'] = self.lcf_ref
        self.lcf_model['rlt'] = xm.lcf(self.lcf_ref_spec, self.norm_spec, constr=self.lcf_constr_use)
        self.lcf_fit = self.lcf_model['rlt'][0].reshape(self.lcf_model['rlt'][0].shape[0], *self.spec.shape[1:])
        self.lcf_fit_err = self.lcf_model['rlt'][1].reshape(*self.spec.shape[1:])

    @msgit(wd=100, fill='+')
    def normalize_xanes(self, eng0, order=1, save_pre_post=False):
        """
        For 2D XANES, self.spec dimensions are [eng, 2D_space]
        For 3D XANES, self.spec dimensions are [eng, 3D_space]

        normalize_xanes includes two steps: 1) pre-edge background subtraction (order = 0)),
                                            2) post-edge fitted or average mu normalization (order = 1)
        return: ndarray, same dimension as spectrum.shape
        """
        eng = deepcopy(self.eng)
        for ii in range(1, self.pre_edge_fit_rlt[0].ndim):
            eng = eng[:, np.newaxis]

        if order == 0:
            self.norm_spec = (self.spec - self.pre_edge_mean_map) / \
                             (self.post_edge_mean_map - self.pre_edge_mean_map)
            self.norm_spec[np.isnan(self.norm_spec)] = 0
            self.norm_spec[np.isinf(self.norm_spec)] = 0
            self.spec_normalized = True
        elif order == 1:
            if save_pre_post:
                e0_idx = xm.index_of(self.eng, eng0)
                self.pre_edge_fit = self.cal_pre_edge_fit(self.eng)
                self.post_edge_fit = self.cal_post_edge_fit([eng])
                self.norm_spec = (self.spec - self.pre_edge_fit) / (
                        self.post_edge_fit[e0_idx] - self.pre_edge_fit[e0_idx])
                self.norm_spec[np.isnan(self.norm_spec)] = 0
                self.norm_spec[np.isinf(self.norm_spec)] = 0
            else:
                self.norm_spec = (self.spec -
                                  self.cal_pre_edge_fit([eng0])) / (
                                         self.cal_post_edge_fit([eng0]) -
                                         self.cal_pre_edge_fit([eng0]))
                self.norm_spec[np.isnan(self.norm_spec)] = 0
                self.norm_spec[np.isinf(self.norm_spec)] = 0
            self.spec_normalized = True
        else:
            print('Order can only be "1" or "0".')
            self.spec_normalized = False

    def remove_sli(self, id_list):
        self.spec = np.delete(self.spec, id_list, axis=0)
        self.eng = np.delete(self.eng, id_list, axis=0)
        self.removed_img_ids = id_list
        self.pre_es_idx = xm.index_of(self.eng, self.pre_es)
        self.pre_ee_idx = xm.index_of(self.eng, self.pre_ee)
        self.post_es_idx = xm.index_of(self.eng, self.post_es)
        self.post_ee_idx = xm.index_of(self.eng, self.post_ee)

    def save_results(self, filename, dtype='2D_XANES', **kwargs):
        filename = Path(filename)
        assert filename.suffix.lower() in ['.h5', '.hdf', '.hdf5'], \
            'unsupported file format...'
        if not filename.parent.exists():
            filename.parent.Path.mkdir(parents=True)

        if dtype == '2D_XANES':
            f = h5py.File(filename, 'a')
            if 'XANES_preprocessing' not in f:
                g1 = f.create_group('XANES_preprocessing')
            else:
                del f['XANES_preprocessing']
                g1 = f.create_group('XANES_preprocessing')

            g1.create_dataset('preset_edge_eng', data=self.preset_edge_eng)
            g1.create_dataset('removed_img_ids', data=self.removed_img_ids)
            g1.create_dataset('pre_edge_fit_coef', data=self.pre_edge_fit_rlt[0])
            g1.create_dataset('post_edge_fit_coef', data=self.post_edge_fit_rlt[0])
            g1.create_dataset('pre_edge_sd_map', data=self.pre_edge_sd_map)
            g1.create_dataset('post_edge_sd_map', data=self.post_edge_sd_map)
            g1.create_dataset('pre_edge_avg', data=self.pre_avg)
            g1.create_dataset('post_edge_avg', data=self.post_avg)
            g1.create_dataset('edge_jump_map', data=self.edge_jump_map)
            g11 = g1.create_group('fitting_configurations')
            g11.create_dataset('pre_edge_energy_range', data=[self.pre_es, self.pre_ee])
            g11.create_dataset('post_edge_energy_range', data=[self.post_es, self.post_ee])

            if 'XANES_filtering' not in f:
                g2 = f.create_group('XANES_filtering')
            else:
                del f['XANES_filtering']
                g2 = f.create_group('XANES_filtering')

            g21 = g2.create_group('edge_jump_filter')
            g21.create_dataset('edge_jump_threshold', data=self.edge_jump_thres)
            g21.create_dataset('edge_jump_mask', data=self.edge_jump_mask.astype(np.int8))
            g22 = g2.create_group('fitted_edge_fitler')
            g22.create_dataset('fitted_edge_threshold', data=self.fitted_edge_thres)
            g22.create_dataset('fitted_edge_mask', data=self.fitted_edge_mask.astype(np.int8))

            if 'XANES_results' not in f:
                g3 = f.create_group('XANES_results')
            else:
                del f['XANES_results']
                g3 = f.create_group('XANES_results')

            g3.create_dataset('eng', data=self.eng.astype(np.float32), dtype=np.float32)
            g3.create_dataset('normalized_spectrum', data=self.norm_spec.astype(np.float32), dtype=np.float32)
            g3.create_dataset('edge_pos', data=self.auto_e0.astype(np.float32), dtype=np.float32)
            g3.create_dataset('edge_0.5_pos', data=self.edge_pos.astype(np.float32), dtype=np.float32)
            g3.create_dataset('whiteline_pos_fit', data=self.wl_pos_fit.astype(np.float32), dtype=np.float32)
            g3.create_dataset('whiteline_pos_direct', data=self.wl_pos_dir.astype(np.float32), dtype=np.float32)

            f.close()
        elif dtype == '3D_XANES':
            f = h5py.File(filename, 'a')
            if 'processed_XANES3D' not in f:
                g1 = f.create_group('processed_XANES3D')
            else:
                del f['processed_XANES3D']
                g1 = f.create_group('processed_XANES3D')

            g1.create_dataset('removed_img_ids',
                              data=self.removed_img_ids)
            g1.create_dataset('pre_edge_fit_coef',
                              data=self.pre_edge_fit_rlt[0])
            g1.create_dataset('post_edge_fit_coef',
                              data=self.post_edge_fit_rlt[0])
            g1.create_dataset('pre_edge_sd_map',
                              data=self.pre_edge_sd_map)
            g1.create_dataset('post_edge_sd_map',
                              data=self.post_edge_sd_map)
            g1.create_dataset('pre_edge_avg',
                              data=self.pre_avg)
            g1.create_dataset('post_edge_avg',
                              data=self.post_avg)
            g1.create_dataset('edge_jump_map',
                              data=self.edge_jump_map)
            g1.create_dataset('edge_jump_threshold',
                              data=self.edge_jump_thres)
            g1.create_dataset('fitted_edge_threshold',
                              data=self.fitted_edge_thres)
            g11 = g1.create_group('fitting_configurations')
            g11.create_dataset('pre_edge_energy_range',
                               data=[self.pre_es, self.pre_ee])
            g11.create_dataset('post_edge_energy_range',
                               data=[self.post_es, self.post_ee])

            if 'XANES_results' not in f:
                g3 = f.create_group('XANES_results')
            else:
                del f['XANES_results']
                g3 = f.create_group('XANES_results')

            g3.create_dataset('eng',
                              data=self.eng.astype(np.float32),
                              dtype=np.float32)
            for key, item in kwargs.items():
                g3.create_dataset(key,
                                  data=item.astype(np.float32),
                                  dtype=np.float32)
            f.close()
        else:
            print('Unrecognized data type!')

    def set_external_mask(self, mask):
        self.ext_mask = mask

    def update_with_auto_e0(self):
        self.find_edge()

        self.pre_ee = self.auto_e0 + (self.pre_ee - self.preset_edge_eng)
        self.pre_ee_idx = xm.index_of(self.eng, self.pre_ee)

        self.post_es = self.auto_e0 + (self.post_es - self.preset_edge_eng)
        self.post_es_idx = xm.index_of(self.eng, self.post_es)
